// Copyright (C) 2017 Ford Motor Company
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#include <QRemoteObjectNode>

#include "rep_pod_replica.h"

int main(int argc, char **argv)
{
  QRemoteObjectNode node;

  MyClassReplica replica;

  return 0;
}
