// Copyright (C) 2017 Ford Motor Company
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#include "speedometer.h"

Speedometer::Speedometer(QObject *parent) :
  SpeedometerSimpleSource(parent)
{
}

Speedometer::~Speedometer()
{
}
